--  pack.e
--  jiri babor
--  jbabor@paradise.net.nz
--  00-01-29
--  version 1.10

----------------------------------------------------------------------
--  Usage:
--  pack(inputFilename, outputFilename)
--  unpack(inputFilename, outputFilename)
----------------------------------------------------------------------

include machine.e
include file.e

object o
sequence b,bits,bo
integer bp,byte,next,fi,fo,len,sp,n

procedure append_bits()
    -- convert flag bits to bytes and append them to end of output
    sequence bits
    integer m
    bits = b > #FF
    m = remainder(length(bits),8)
    if m then
        m = 8-m
        bits &= repeat(0, m)
    end if
    for i=1 to length(bits) by 8 do
        b &= bits_to_int(bits[i..i+7])
    end for
end procedure

global procedure pack(
    sequence inFile,    -- name of file to be compressed
    sequence outFile    -- name of compressed output file
    )
    fi=open(inFile, "rb")
    if fi = -1 then
        puts(1, "Input file " & inFile & " not found!")
        abort(1)
    end if

    o=seek(fi,-1)           -- go to end of input file
    len=where(fi)           -- get length of input file in bytes
    o=seek(fi,0)            -- go back to beginning of input file
    b={}                    -- processed (compressed) data sequence
    bp=0                    -- init data pointer
    sp=1                    -- init slider pointer
    next=getc(fi)           -- get very first byte
    while next!=-1 do       -- while not end of input file
        while sp<bp-1 do    -- shrink routine
            if b[sp]=b[bp] then
                if b[sp+1]=next then
                    next=and_bits(sp-1,#FF)+#100
                    bp-=1
                    b=b[1..bp]
                end if
            end if
            sp+=1           -- update search zone pointer
        end while
        b=append(b,next)    -- add to output
        bp+=1               -- update data pointer
        next=getc(fi)       -- retrive the next byte from input file
        sp=bp-#101          -- go back to beginning of search zone
        if sp<1 then        -- but not past the beginning of file
            sp=1
        end if
    end while
    close(fi)               -- close input file

    append_bits()           -- add flags to output
    fo=open(outFile,"wb")   -- open output file
    puts(fo,b)              -- write out
    close(fo)               -- close output file
end procedure

procedure Unpack()
    integer mp,mq,p,q
    p=n                         -- initial length of b, bits
    bo={}                       -- output sequence
    while p do
        if bits[p] then
            mp=and_bits(p,#FF)
            mq=b[p]
            q=p-mp+mq+1
            while q>p-2 do q -= #100 end while
            b[p]=b[q]
            bits[p]=bits[q]
            b &= b[q+1]
            bits &= bits[q+1]
            p+=1
        else
            bo=prepend(bo,b[p])
            p-=1
            b=b[1..p]
            bits=bits[1..p]
        end if
    end while
end procedure -- Unpack

global procedure unpack(
    sequence inFile,
    sequence outFile
    )

    fi=open(inFile, "rb")
    if fi=-1 then
        puts(1, "Input file " & inFile & " not found!")
        abort(1)
    end if

    -- load
    b={}                        -- input (compressed) bytes sequence
    byte=getc(fi)
    while byte!=-1 do
        b&=byte
        byte=getc(fi)
    end while
    close(fi)

    -- convert 'flag' bytes to bits
    n=floor(length(b)*8/9)      -- number of 'data' bytes
    bits={}
    for i=n+1 to length(b) do
        bits&=int_to_bits(b[i],8)
    end for

    -- trim sequences
    b=b[1..n]
    bits=bits[1..n]

    Unpack()

    fo=open(outFile, "wb")      -- write out
    puts(fo, bo)
    close(fo)
end procedure
