--  unpack.ex : restore data compressed by pack.ex
--  jiri babor
--  jbabor@paradise.net.nz
--  00-01-29
--  version 1.10
----------------------------------------------------------------------
--  usage:  ex  unpack inFilename  outFilename
--  where   inFilename : file of data compressed using pack.ex
--          outFilename: output file with de-compressed data
----------------------------------------------------------------------

include machine.e

sequence b,bits,bo,cl
atom t
integer byte,fi,fo,n

procedure help(sequence error_message)
    puts(1,"Error :  " & error_message & "\n")
    puts(1,"Syntax:  ex  unpack  inFilename  outFilename\n")
    abort(1)
end procedure

procedure unpack()
    integer mp,mq,p,q
    p=n                         -- initial length of b, bits
    bo={}                       -- output sequence
    while p do
        if bits[p] then
            mp=and_bits(p,#FF)
            mq=b[p]
            q=p-mp+mq+1
            while q>p-2 do q -= #100 end while
            b[p]=b[q]
            bits[p]=bits[q]
            b&=b[q+1]
            bits&=bits[q+1]
            p+=1
        else
            bo=prepend(bo,b[p])
            p-=1
            b=b[1..p]
            bits=bits[1..p]
        end if
    end while
end procedure

-- main ------------------------------------------------------------------------

cl=command_line()
if length(cl)!=4 then
    help("Incorrect number of arguments!")
end if
fi=open(cl[3],"rb")
if fi=-1 then
    help("Input file " & cl[3] & " not found!")
end if

tick_rate(100)                  -- better timer resolution
-- load
t=time()
b={}                            -- input (compressed) bytes sequence
byte=getc(fi)
while byte!=-1 do
    b &= byte
    byte=getc(fi)
end while
close(fi)

-- convert 'flag' bytes to bits
n=floor(length(b)*8/9)          -- number of 'data' bytes
bits={}
for i=n+1 to length(b) do
    bits &= int_to_bits(b[i],8)
end for

-- trim sequences
b = b[1..n]
bits = bits[1..n]

unpack()

fo=open(cl[4],"wb")             -- write out
puts(fo, bo)
close(fo)
t=time()-t

printf(1,"De-compression time  %.2f  seconds\n",t)
